/*****************************************************************************

 httpclient.d:

 HTTP Client

 Copyright (c) 2005-2008 Yoshinori SUDA.

 This program is free software; you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation; either version 2 of the License, or 
 (at your option) any later version.

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 merchantability or fitness for a particular purpose. 

 See gpl.txt for more details.

 ----------------------------------------------------------------------------

 Requirement:

 - wsock32.lib

 ----------------------------------------------------------------------------

 Public Functions:

 - getHttpData
 - getHttpText

 ----------------------------------------------------------------------------

 Class Hierarchy:

 - HttpException : Exception

*****************************************************************************/

module suda.httpclient;

private import std.conv;
private import std.socket, std.socketstream;
private import std.string;
private import std.uri;
private import suda.ver;
private import suda.jcode;

enum Charset { UTF_8, EUC_JP, SHIFT_JIS };

public const char[] SUDA_USER_AGENT = "suda.httpclient/" ~ SUDA_VERSION_STRING;

class HttpException : Exception
{
	this(char[] err_msg)
	{
		super(err_msg);
	}
}

// url から読み出し、ヘッダを head に、内容を data に返す
void getHttpData(char[] url, out char[] head, out ubyte[] data)
{
	__get_http_data(__url_encode(url), head, data);
}

// url から対象をテキストページと仮定して読み出し、結果を UTF-8 文字列で返す
// ヘッダで文字コードが指定されていなければ charset であると仮定される
char[] getHttpText(char[] url, Charset charset = Charset.UTF_8)
{
	char[] head;
	ubyte[] data;
	char[][] headlines;
	
	__get_http_data(__url_encode(url), head, data);
	headlines = splitlines(head);
	
	foreach (char[] s; headlines) {
		if (ifind(s, "Content-Type") >= 0) {
			if (ifind(s, "euc-jp") >= 0)
				charset = Charset.EUC_JP;
			else if (ifind(s, "shift_jis") >= 0)
				charset = Charset.SHIFT_JIS;
			else if (ifind(s, "utf-8") >= 0)
				charset = Charset.UTF_8;
		}
	}
	switch (charset) {
	case Charset.EUC_JP:
		return euc_to_utf8(cast(char[])data);
	case Charset.SHIFT_JIS:
		return sjis_to_utf8(cast(char[])data);
	default:
		return cast(char[])data;
	}
}

private char[] __url_encode(char[] url)
{
	url = strip(url);
	url = std.uri.encode(url);
/*
	url = replace(url, "&", "%26");
	url = replace(url, "?", "%3f");
*/
	return url;
}

private const int READ_BUFFER_SIZE = 2048;

private void __get_http_data(char[] url, out char[] head, out ubyte[] data)
{
	int i;
	char[] domain;
	uint port = 80;
	auto Socket sock;
	SocketStream client;
	char[] line;
	ubyte[READ_BUFFER_SIZE] buf;
	
	if ((i = find(url, "://")) >= 0) {
		if (url[0..7] != "http://")
			throw new HttpException("'http://' expected");
		url = url[7..length];
	}
	if ((i = find(url, '#')) >= 0)	// anchor
		url = url[0..i];
	if ((i = find(url, "/")) >= 0) {
		domain = url[0..i];
		url = url[i..length];
	} else {
		domain = url;
		url = "/";
	}
	if ((i = find(domain, ":")) >= 0) {
		port = toUshort(domain[(i+1)..length]);
		domain = domain[0..i];
	}
	
	sock = new TcpSocket(new InternetAddress(domain, port));
	client = new SocketStream(sock);
	
	if(port != 80)
		domain = format("%s:%d", domain, port);
	client.writeString(
		"GET " ~ url ~ " HTTP/1.1\r\n"
		"Host: " ~ domain ~ "\r\n"
		"User-Agent: " ~ SUDA_USER_AGENT ~ "\r\n"
		"\r\n");
	
	do {
		line = client.readLine();
		head ~= line;
		head ~= "\n";
	} while (line.length != 0);
	
	while(!client.eof()) {
		i = client.readBlock(&buf, buf.length);
		data ~= buf[0..i];
	}
}
