/*****************************************************************************

 smf.d:

 Classes for Generating Standard MIDI Files.

 Copyright (c) 2005 Yoshinori SUDA. All rights reserved.

 This program is free software; you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation; either version 2 of the License, or 
 (at your option) any later version.

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 merchantability or fitness for a particular purpose. 

 See gpl.txt for more details.

 ----------------------------------------------------------------------------

 Class Hierarchy:

 - MidiData
    - MidiSimpleData
       - MidiEvent
          - MidiNoteOff
          - MidiNoteOn
          - MidiPolyphonicKeyPressure
          - MidiControlChange
          - MidiProgramChange
          - MidiChannelPressure
          - MidiPitchBend
       - MidiSysEx
       - MidiSysExVerbatim
       - MidiMetaEvent
          - MidiSequenceNumber (reserved)
          - MidiText
          - MidiCopyright
          - MidiSequenceName
          - MidiInstrumentName
          - MidiLyric
          - MidiMarker
          - MidiQuePoint
          - MidiProgramName
          - MidiDeviceName
          - MidiChannelPrefix
          - MidiPort
          - MidiEndOfTrack
          - MidiTempo
          - MidiSmpteOffset (reserved)
          - MidiTimeSignature
          - MidiKeySignature (reserved)
          - MidiSequencerSpecific (reserved)
    - MidiComplexData
    - MidiDataChain
       - MidiTrack
    - MidiFile

*****************************************************************************/

module suda.smf;

abstract class MidiData
{
	// SMFデータ形式に変換
	ubyte[] serialize() { return null; };
	
	// 整数を範囲内に制限
	static int normalize(int target, int min, int max)
	{
		if (target < min)
			return min;
		else if (target > max)
			return max;
		else
			return target;
	}
	
	// デルタ・タイムをSMFの可変長形式に変換
	static ubyte[] serializeDeltaTime(int ticks)
	in {
		assert(0 <= ticks && ticks <= 0xfffffff);
	}
	body {
		int l = 0;
		ubyte[] r = new ubyte[4];
		
		ticks = normalize(ticks, 0, 0xfffffff);
		if (ticks >= 0b1_0000000_0000000_0000000)
			r[l++] = ticks >> (7 * 3) & 0xff | 0x80;
		if (ticks >= 0b1_0000000_0000000)
			r[l++] = ticks >> (7 * 2) & 0xff | 0x80;
		if (ticks >= 0b1_0000000)
			r[l++] = ticks >> 7 & 0xff | 0x80;
		r[l++] = ticks & 0x7f;
		r.length = l;
		return r;
	}
	
	// serialize() の結果を表示する（動作テスト用）
	debug void _print()
	{
		ubyte[] r = serialize();
		
		for (int i; i < r.length; i++)
			printf("%02X ", r[i]);
		printf("\n");
	}
}

// MidiSimpleData の serialize はデルタ・タイムを含まない
class MidiSimpleData : MidiData
{
	// そのまま出力するかどうかのフラグ。
	// ture の場合、省略や同等のデータに変更される可能性がある。
	// 例えば、ランニング・ステータス、ノート・オフをノート・オンに、
	// 同時的なコントロール・チェンジを一つにまとめるなど。
	bit omissible = true;
	
	// serialize() 結果の大きさを返す抽象関数
	// serialize().length と同等だが高速である
	int size() { return 0; };
}

class MidiEvent : MidiSimpleData
{
private:
	int channel_number;
	invariant {
		assert(0 <= channel_number && channel_number <= 15);
	}
public:
	// チャンネル番号のプロパティ
	int channel()
	{
		return channel_number;
	}
	int channel(int number)
	{
		return channel_number = normalize(number, 0, 15);
	}
}

class MidiNoteOff : MidiEvent
{
private:
	int note_num;
	int velocity_num;
public:
	this()
	{ 
		// empty
	}
	this(int channel, int note, int velocity)
	{
		super.channel = channel;
		note_num = normalize(note, 0, 0x7f);
		velocity_num = normalize(velocity, 0, 0x7f);
	}
	
	int note()
	{
		return note_num;
	}
	int note(int number)
	{
		return note_num = normalize(number, 0, 0x7f);
	}
	
	int velocity()
	{
		return velocity_num;
	}
	int velocity(int number)
	{
		return velocity_num = normalize(number, 0, 0x7f);
	}
	
	override ubyte[] serialize()
	{
		ubyte[] r = new ubyte[3];
		
		r[0] = 0x80 + super.channel;
		r[1] = note_num;
		r[2] = velocity_num;
		return r;
	}
	
	override int size()
	{
		return 3;
	}
}

class MidiNoteOn : MidiEvent
{
private:
	int note_num;
	int velocity_num;
public:
	this()
	{ 
		// empty
	}
	this(int channel, int note, int velocity)
	{
		super.channel = channel;
		note_num = normalize(note, 0, 0x7f);
		velocity_num = normalize(velocity, 0, 0x7f);
	}
	
	int note()
	{
		return note_num;
	}
	int note(int number)
	{
		return note_num = normalize(number, 0, 0x7f);
	}
	
	int velocity()
	{
		return velocity_num;
	}
	int velocity(int number)
	{
		return velocity_num = normalize(number, 0, 0x7f);
	}
	
	override ubyte[] serialize()
	{
		ubyte[] r = new ubyte[3];
		
		r[0] = 0x90 + super.channel;
		r[1] = note_num;
		r[2] = velocity_num;
		return r;
	}
	
	override int size()
	{
		return 3;
	}
}

class MidiPolyphonicKeyPressure : MidiEvent
{
private:
	int note_num;
	int velocity_num;
public:
	this()
	{ 
		// empty
	}
	this(int channel, int note, int velocity)
	{
		super.channel = channel;
		note_num = normalize(note, 0, 0x7f);
		velocity_num = normalize(velocity, 0, 0x7f);
	}
	
	int note()
	{
		return note_num;
	}
	int note(int number)
	{
		return note_num = normalize(number, 0, 0x7f);
	}
	
	int velocity()
	{
		return velocity_num;
	}
	int velocity(int number)
	{
		return velocity_num = normalize(number, 0, 0x7f);
	}
	
	override ubyte[] serialize()
	{
		ubyte[] r = new ubyte[3];
		
		r[0] = 0xA0 + super.channel;
		r[1] = note_num;
		r[2] = velocity_num;
		return r;
	}
	
	override int size()
	{
		return 3;
	}
}

class MidiControlChange : MidiEvent
{
private:
	int c_num;
	int v_num;
public:
	enum {
		BANK_SELECT_MSB = 0, BANK_SELECT_LSB = 32, 
		MODULATION = 1, PORTAMENTO_TIME = 5, 
		VOLUME = 7, PANPOT = 10, EXPRESSION = 11, 
		HOLD1 = 64, PORTAMENTO_ON_OFF, SOSTENUTO, SOFT, 
		FILTER_RESONANCE = 71, RELEASE_TIME, ATTACK_TIME, BRIGHTNESS, 
		DECAY_TIME, VIBRATO_RATE, VIBRATO_DEPTH, VIBRATO_DELAY, 
		REVERB_SEND_LEVEL = 91, CHORUS_SEND_LEVEL = 93, 
		DATA_ENTRY_MSB = 6, DATA_ENTRY_LSB = 38, 
		NRPN_LSB = 98, NRPN_MSB, RPN_LSB, RPN_MSB
	}
	
	this()
	{ 
		// empty
	}
	this(int channel, int controller, int value)
	{
		super.channel = channel;
		c_num = normalize(controller, 0, 0x7f);
		v_num = normalize(value, 0, 0x7f);
	}
	
	int controller()
	{
		return c_num;
	}
	int controller(int number)
	{
		return c_num = normalize(number, 0, 0x7f);
	}
	
	int value()
	{
		return v_num;
	}
	int value(int number)
	{
		return v_num = normalize(number, 0, 0x7f);
	}
	
	override ubyte[] serialize()
	{
		ubyte[] r = new ubyte[3];
		
		r[0] = 0xB0 + super.channel;
		r[1] = c_num;
		r[2] = v_num;
		return r;
	}
	
	override int size()
	{
		return 3;
	}
}

class MidiProgramChange : MidiEvent
{
private:
	int prog_num;
public:
	this()
	{ 
		// empty
	}
	this(int channel, int program)
	{
		super.channel = channel;
		prog_num = normalize(program, 0, 0x7f);
	}
	
	int program()
	{
		return prog_num;
	}
	int program(int number)
	{
		return prog_num = normalize(number, 0, 0x7f);
	}
	
	override ubyte[] serialize()
	{
		ubyte[] r = new ubyte[2];
		
		r[0] = 0xC0 + super.channel;
		r[1] = prog_num;
		return r;
	}
	
	override int size()
	{
		return 2;
	}
}

class MidiChannelPressure : MidiEvent
{
private:
	int velocity_num;
public:
	this()
	{ 
		// empty
	}
	this(int channel, int velocity)
	{
		super.channel = channel;
		velocity_num = normalize(velocity, 0, 0x7f);
	}
	
	int velocity()
	{
		return velocity_num;
	}
	int velocity(int number)
	{
		return velocity_num = normalize(number, 0, 0x7f);
	}
	
	override ubyte[] serialize()
	{
		ubyte[] r = new ubyte[2];
		
		r[0] = 0xD0 + super.channel;
		r[1] = velocity_num;
		return r;
	}
	
	override int size()
	{
		return 2;
	}
}

class MidiPitchBend : MidiEvent
{
private:
	int pitch_num;
public:
	this()
	{ 
		// empty
	}
	this(int channel, int pitch)
	{
		super.channel = channel;
		pitch_num = normalize(pitch, -0x2000, 0x1fff);
	}
	
	int pitch()
	{
		return pitch_num;
	}
	int pitch(int number)
	{
		return pitch_num = normalize(number, -0x2000, 0x1fff);
	}
	
	override ubyte[] serialize()
	{
		ubyte[] r = new ubyte[3];
		int bender = pitch_num + 0x2000;
		
		r[0] = 0xE0 + super.channel;
		r[1] = bender & 0x7f;
		r[2] = bender >> 7;
		return r;
	}
	
	override int size()
	{
		return 3;
	}
}

class MidiSysEx : MidiSimpleData
{
private:
	ubyte[] data_msg;
public:
	this()
	{
		super.omissible = false;
	}
	this(ubyte[] data)
	{
		super.omissible = false;
		data_msg = data;
	}
	
	ubyte[] data()
	{
		return data_msg;
	}
	ubyte[] data(ubyte[] msg)
	{
		return data_msg = msg;
	}
	
	override ubyte[] serialize()
	{
		ubyte[1] r1 = 0xf0;
		ubyte[1] r2 = 0xf7;
		
		return r1 ~ serializeDeltaTime(data_msg.length + 1) ~ data_msg ~ r2;
	}
	
	override int size()
	{
		return 1 + serializeDeltaTime(data_msg.length + 1).length 
			+ data_msg.length + 1;
	}
}

// 0xf7 で始まる SysEx （非推奨）
class MidiSysExVerbatim : MidiSimpleData
{
private:
	ubyte[] data_msg;
public:
	this()
	{
		super.omissible = false;
	}
	this(ubyte[] data)
	{
		super.omissible = false;
		data_msg = data;
	}
	
	ubyte[] data()
	{
		return data_msg;
	}
	ubyte[] data(ubyte[] msg)
	{
		return data_msg = msg;
	}
	
	override ubyte[] serialize()
	{
		ubyte[1] r1 = 0xf7;
		
		return r1 ~ serializeDeltaTime(data_msg.length) ~ data_msg;
	}
	
	override int size()
	{
		return 1 + serializeDeltaTime(data_msg.length).length + data_msg.length;
	}
}

class MidiMetaEvent : MidiSimpleData
{
private:
	ubyte event_num;
	ubyte[] data_msg;
public:
	this(ubyte event, ubyte[] data)
	{
		event_num = event;
		data_msg = data;
	}
	
	ubyte event()
	{
		return event_num;
	}
	
	ubyte[] data()
	{
		return data_msg;
	}
	
	override ubyte[] serialize()
	{
		ubyte[] r = new ubyte[2];
		
		r[0] = 0xff;
		r[1] = event_num;
		return r ~ serializeDeltaTime(data_msg.length) ~ data_msg;
	}
	
	override int size()
	{
		return 2 + serializeDeltaTime(data_msg.length).length + data_msg.length;
	}
}

/**** 未実装
class MidiSequenceNumber : MidiMetaEvent
{
}
****/

class MidiText : MidiMetaEvent
{
	this(ubyte[] text)
	{
		super(0x01, text);
	}
}

class MidiCopyright : MidiMetaEvent
{
	this(ubyte[] text)
	{
		super(0x02, text);
	}
}

class MidiSequenceName : MidiMetaEvent
{
	this(ubyte[] text)
	{
		super(0x03, text);
	}
}

class MidiInstrumentName : MidiMetaEvent
{
	this(ubyte[] text)
	{
		super(0x04, text);
	}
}

class MidiLyric : MidiMetaEvent
{
	this(ubyte[] text)
	{
		super(0x05, text);
	}
}

class MidiMarker : MidiMetaEvent
{
	this(ubyte[] text)
	{
		super(0x06, text);
	}
}

class MidiQuePoint : MidiMetaEvent
{
	this(ubyte[] text)
	{
		super(0x07, text);
	}
}

class MidiProgramName : MidiMetaEvent
{
	this(ubyte[] text)
	{
		super(0x08, text);
	}
}

class MidiDeviceName : MidiMetaEvent
{
	this(ubyte[] text)
	{
		super(0x09, text);
	}
}

class MidiChannelPrefix : MidiMetaEvent
{
	this(int channel)
	{
		ubyte[] r = new ubyte[1];
		
		r[0] = normalize(channel, 0, 15);
		super(0x20, r);
	}
}

// SMF v1.0 規格外
class MidiPort : MidiMetaEvent
{
	this(int port)
	{
		ubyte[] r = new ubyte[1];
		
		r[0] = normalize(port, 0, 0x7f);
		super(0x21, r);
	}
}

// MidiTrack 等では自動的に挿入される
class MidiEndOfTrack : MidiMetaEvent
{
	this()
	{
		super(0x2f, new ubyte[0]);
	}
	
	override ubyte[] serialize()
	{
		ubyte[] r = new ubyte[3];
		
		r[0] = 0xff;
		r[1] = 0x2f;
		r[2] = 0x00;
		return r;
	}
	
	override int size()
	{
		return 3;
	}
}

// テンポをBPMで指定
class MidiTempo : MidiMetaEvent
{
	this(int bpm)
	{
		// int t = 60 * 1000000 / bpm;
		int t = (60 * 1000000 + (bpm - 1)) / bpm;
		ubyte[] r = new ubyte[3];
		
		r[0] = t >> 16 & 0xff;
		r[1] = t >>  8 & 0xff;
		r[2] = t       & 0xff;
		super(0x51, r);
	}
	
	override int size()
	{
		return 6;
	}
}

/**** 未実装
class MidiSmpteOffset : MidiMetaEvent
{
}
****/

class MidiTimeSignature : MidiMetaEvent
{
	this(int numerator, int denominator)
	{
		int d, a;
		ubyte[] r = new ubyte[4];
		
		numerator = normalize(numerator, 1, 0x7f);
		denominator = normalize(denominator, 1, 0x7f);
		for (d = 1, a = 2; d <= 0x7f; d++) {
			if (denominator <= a)
				break;
			a *= 2;
		}
		r[0] = numerator;
		r[1] = d;
		r[2] = 24;
		r[3] = 8;
		super(0x58, r);
	}
}

/**** 未実装
class MidiKeySignature : MidiMetaEvent
{
}

class MidiSequencerSpecific : MidiMetaEvent
{
}
****/

// 同時的な複数の MidiSimpleData をまとめるクラス
// serialize() の先頭にデルタ・タイムは含まないが、
// MidiSimpleData 同士の接合部部分にはゼロが含まれる
class MidiComplexData : MidiData
{
private:
	// 単方向リスト
	struct Item {
		MidiSimpleData data;
		Item* next;
	}
	Item* head = null;
	int items = 0;
public:
	int length()
	{
		return items;
	}
	
	// 位置 index に対応する要素を返す
	MidiSimpleData get(int index)
	{
		Item* p = head;
		
		if (index < 0 || items <= index)
			throw new Exception("out of range in MidiComplexData.get()");
		for (int i; i < index; i++)
			p = p.next;
		return p.data;
	}
	MidiSimpleData opIndex(int index)
	{
		return get(index);
	}
	
	// 位置 index の要素を d と置換する
	MidiSimpleData set(int index, MidiSimpleData d)
	{
		Item* p = head;
		
		if (index < 0 || items <= index)
			throw new Exception("out of range in MidiComplexData.set()");
		for (int i; i < index; i++)
			p = p.next;
		return p.data = d;
	}
	MidiSimpleData opIndexAssign(MidiSimpleData d, int index)
	{
		return set(index, d);
	}
	
	// 最後尾に連結し、連結位置を返す
	int add(MidiSimpleData d)
	{
		int i;
		Item dummy;
		Item* p;
		
		dummy.next = head;
		p = &dummy;
		while (p.next != null) {
			p = p.next;
			i++;
		}
		p.next = new Item;
		p.next.data = d;
		p.next.next = null;
		if (head == null)
			head = p.next;
		items++;
		return i;
	}
	
	int add(MidiComplexData d)
	{
		int i;
		Item* p;
		
		if (head == null) {
			head = d.head;
			items = d.items;
			return 0;
		}
		p = head;
		while (p.next != null) {
			p = p.next;
			i++;
		}
		p.next = d.head;
		items += d.items;
		return i;
	}
	
	// 先頭に連結する、連結位置（常にゼロ）を返す
	int addInFront(MidiSimpleData d)
	{
		Item* p = new Item;
		
		p.data = d;
		p.next = head;
		head = p;
		items++;
		return 0;
	}
	
	int addInFront(MidiComplexData d)
	{
		d.add(this);
		head = d.head;
		items = d.items;
		return 0;
	}
	
	override ubyte[] serialize()
	{
		const ubyte[1] delta_zero = [ 0x00 ];
		ubyte[] r;
		Item* p;
		
		if (head == null)
			return r;
		r = head.data.serialize();
		p = head.next;
		while (p != null) {
			r ~= delta_zero;
			r ~= p.data.serialize();
			p = p.next;
		}
		return r;
	}
	
	bit[] getOmissibleMap()
	out(r) {
		debug assert(r.length == serialize().length);
	}
	body {
		bit[] r;
		Item* p;
		int l, i, sz;
		
		if (head == null)
			return r;
		p = head;
		l = p.data.size();
		while (p.next != null) {
			p = p.next;
			l += (1 + p.data.size());
		}
		r = new bit[l];
		
		p = head;
		sz = p.data.size();
		for (i = 0; i < sz; i++)
			r[i] = p.data.omissible;
		while (p.next != null) {
			p = p.next;
			sz = (1 + p.data.size());
			for (; i < sz; i++)
				r[i] = p.data.omissible;
		}
		return r;
	}
}

// 時間差のある MidiSimpleData の集まり
// 時間位置は tick 数、つまり先頭からのデルタ・タイムの合計で表す
// （実装上はキーを tick 数とする MidiComplexData の連想配列）
// serialize() 結果の先頭にもデルタ・タイムが含まれる
// tick 数には負の数も指定できる。連結時の相対用。serialize() では無視
class MidiDataChain : MidiData
{
private:
	MidiComplexData[int] chain;
	
	MidiComplexData validate(int tick)
	{
		if (tick in chain)
			return chain[tick];
		else
			return chain[tick] = new MidiComplexData();
	}
public:
	bool isValid(int tick)
	{
		if (tick in chain)
			return true;
		else
			return false;
	}
	
	int add(int tick, MidiSimpleData data)
	{
		return validate(tick).add(data);
	}
	
	int addInFront(int tick, MidiSimpleData data)
	{
		return validate(tick).addInFront(data);
	}
	
	int add(int tick, MidiComplexData data)
	{
		if (tick in chain)
			return chain[tick].add(data);
		else
			{ chain[tick] = data; return 0; }
	}
	
	int addInFront(int tick, MidiComplexData data)
	{
		if (tick in chain)
			return chain[tick].addInFront(data);
		else
			{ chain[tick] = data; return 0; }
	}
	
	int add(int tick, MidiDataChain data)
	{
		foreach (int i, MidiComplexData d; data.chain)
			add(tick + i, d);
		return tick;
	}
	
	int addInFront(int tick, MidiDataChain data)
	{
		foreach (int i, MidiComplexData d; data.chain)
			addInFront(tick + i, d);
		return tick;
	}
	
	int mix(MidiDataChain data)
	{
		return add(0, data);
	}
	
	MidiComplexData get(int tick)
	{
		return validate(tick);
	}
	
	MidiComplexData erase(int tick)
	{
		MidiComplexData r;
		
		if (tick in chain) {
			r = chain[tick];
			chain.remove(tick);
			return r;
		} else
			return null;
	}
	
	MidiComplexData replace(int tick, MidiComplexData data)
	{
		return chain[tick] = data;
	}
	
	override ubyte[] serialize()
	{
		int[] ticks = chain.keys;
		ubyte[] result;
		int past = 0;
		
		ticks.sort;
		for (int i = 0; i < ticks.length; i++) {
			int t = ticks[i];
			if (t < 0)
				continue;
			if (chain[t].length == 0)
				continue;
			result ~= serializeDeltaTime(t - past);
			result ~= chain[t].serialize();
			past = t;
		}
		return result;
	}
	
	bit[] getOmissibleMap()
	{
		int[] ticks = chain.keys;
		bit[] result;
		int past = 0;
		
		ticks.sort;
		for (int i; i < ticks.length; i++) {
			int t = ticks[i];
			if (t < 0)
				continue;
			if (chain[t].length == 0)
				continue;
			if (t - past <= 0x7f)
				result ~= new bit[1];
			else
				result ~= new bit[serializeDeltaTime(t - past).length];
			result ~= chain[t].getOmissibleMap();
			past = t;
		}
		return result;
	}
	
	int steps()
	{
		int[] k = chain.keys;
		int max = 0;
		
		foreach (int i; k)
			if (i > max)
				max = i;
		return max;
	}
}

// トラック・データ（MTrkイベント）
// serialize() 結果は "MTrk" に始まり、
// 終端に MidiEndOfTrack が自動付加される。
class MidiTrack : MidiDataChain
{
	override ubyte[] serialize()
	{
		ubyte[] head;
		ubyte[] result;
		uint size;
		
		head = new ubyte[8];
		head[0] = 0x4d;	// 'M'
		head[1] = 0x54;	// 'T'
		head[2] = 0x72;	// 'r'
		head[3] = 0x6b;	// 'k'
		
		result = super.serialize();
		result ~= serializeDeltaTime(0);
		result ~= (new MidiEndOfTrack).serialize();
		
		size = result.length;
		head[4] = size >> (8 * 3) & 0xff;
		head[5] = size >> (8 * 2) & 0xff;
		head[6] = size >>  8      & 0xff;
		head[7] = size            & 0xff;
		
		head ~= result;
		return head;
	}
	
	override bit[] getOmissibleMap()
	{
		bit[] result = new bit[8];
		
		result = super.getOmissibleMap();
		result ~= new bit[1 + (new MidiEndOfTrack).size()];
		
		return result;
	}
	
	// 最適化処理を行う serialize()。
	// ランニング・ステータスの適用、重複していて不用なイベントの削除など
	ubyte[] optimizingSerialization()
	{
		// 未実装
		pragma(msg, "*** Not implemented suda.smf.MidiTrack.optimizingSerialization() yet");
		
		return serialize();
	}
}

// serialize() 結果はそのままSMF形式
// デフォルトではフォーマット1
// toFormat0() でフォーマット0に変換する
class MidiFile : MidiData
{
protected:
	// そのままの順序でSMF化
	// 通常 tracks[0] が Conductor Track
	// tracks[1..16] がそれぞれチャンネル 0..15 に割り当てられる
	MidiTrack[] tracks;
	
	int format = 1;
	int timebase = 48;
public:
	this(MidiTrack[] midi_tracks)
	{
		tracks = midi_tracks;
	}
	this(MidiTrack[] midi_tracks, int midi_timebase)
	{
		tracks = midi_tracks;
		timebase = midi_timebase;
	}
	
	void toFormat0()
	{
		format = 0;
		if (tracks.length <= 1)
			return;
		for (int t = 1; t < tracks.length; t++)
			tracks[0].mix(tracks[t]);
		tracks.length = 1;
	}
	
	override ubyte[] serialize()
	{
		ubyte[] result = new ubyte[14];
		
		result[0] = 0x4d;	// 'M'
		result[1] = 0x54;	// 'T'
		result[2] = 0x68;	// 'h'
		result[3] = 0x64;	// 'd'
		
		result[4] = 0;
		result[5] = 0;
		result[6] = 0;
		result[7] = 6;
		
		result[8] = 0;
		result[9] = format;
		
		result[10] = tracks.length >> 8;
		result[11] = tracks.length & 0xff;
		
		result[12] = timebase >> 8 & 0x7f;
		result[13] = timebase & 0xff;
		
		for (int t; t < tracks.length; t++)
			result ~= tracks[t].optimizingSerialization();
		
		return result;
	}
}
