/*****************************************************************************

 std.d:

 General Utility Functions and Classes

 Copyright (c) 2005-2009 Yoshinori SUDA.

 This program is free software; you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation; either version 2 of the License, or 
 (at your option) any later version.

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 merchantability or fitness for a particular purpose. 

 See gpl.txt for more details.

 ----------------------------------------------------------------------------

 Public Functions:

 - (No Entry)

 ----------------------------------------------------------------------------

 Class Hierarchy:

 - Stack(T)
 - Queue(T)
 - Reader(T)

*****************************************************************************/

module suda.std;

// For Stack Class
class EmptyStackException : Exception
{
	this()
	{
		super("Stack Underflow");
	}
}

// Stack
class Stack(T)
{
private:
	T[] stack;
	int p;

public:
	this(int buffer_size = 64)
	{
		stack.length = buffer_size;
		p = 0;
	}
	
	bool empty()
	{
		return p <= 0;
	}
	
	T peek()
	{
		if (empty)
			throw new EmptyStackException();
		else
			return stack[p - 1];
	}
	
	T pop()
	{
		if (empty)
			throw new EmptyStackException();
		else
			return stack[--p];
	}
	
	T push(T d)
	{
		if (p >= stack.length)
			stack.length = stack.length * 2;
		return stack[p++] = d;
	}
	
	T[] push(T[] a)
	{
		foreach (T d; a)
			push(d);
		return a;
	}
	
	T[] toArray()
	{
		return stack[0..p];
	}
	
	void clear()
	{
		p = 0;
	}
	
	int length()
	{
		return p;
	}
	
	T opIndex(int index)
	{
		if (index < 0 || p <= index)
			throw new Exception("Array Bounds Error");
		return stack[index];
	}
	
	T opIndexAssign(T d, int index)
	{
		if (index < 0 || p <= index)
			throw new Exception("Array Bounds Error");
		return stack[index] = d;
	}
}

// For Queue Class
class EmptyQueueException : Exception
{
	this()
	{
		super("Queue Has No Element");
	}
}

// Queue (FIFO Buffer)
class Queue(T)
{
private:
	T[] queue;
	int o;		// 次に書き込むべき位置
	int p;		// 次に読み込むべき位置

protected:
	void extend()
	{
		T[] twice;
		int i = 0;
		
		twice.length = queue.length * 2;
		while (!empty)
			twice[i++] = poll();
		p = 0;
		o = i;
		queue = twice;
	}

public:
	this(int buffer_size = 64)
	{
		queue.length = buffer_size;
		o = p = 0;
	}
	
	bool empty()
	{
		return o == p;
	}
	
	T poll()
	{
		T r;
		
		if (empty)
			throw new EmptyQueueException();
		r = queue[p++];
		if (p == queue.length)
			p = 0;
		return r;
	}
	
	T peek()
	{
		if (empty)
			throw new EmptyQueueException();
		else
			return queue[p];
	}
	
	T offer(T d)
	{
		int next;
		
		next = o + 1;
		if (next >= queue.length)
			next = 0;
		if (next == p)
			{ extend(); next = o + 1; }
		
		queue[o] = d;
		o = next;
		return d;
	}
	
	T[] offer(T[] a)
	{
		foreach (T d; a)
			offer(d);
		return a;
	}
	
	void clear()
	{
		o = p = 0;
	}
}

// For Reader Class
class EmptyReaderException : Exception
{
	this()
	{
		super("Reader Has No Element");
	}
}

// Reader
class Reader(T)
{
private:
	T[] reader;
	int p;		// 次に読み込むべき位置

public:
	this(T[] buf)
	{
		reader = buf;
		p = 0;
	}
	
	bool empty()
	{
		return p >= buf.length;
	}
	
	T poll()
	{
		if (empty)
			throw new EmptyReaderException();
		return reader[p++];
	}
	
	T peek()
	{
		if (empty)
			throw new EmptyReaderException();
		else
			return reader[p];
	}
	
	void clear()
	{
		reader = null;
		p = 0;
	}
	
	int length()
	{
		return reader.length - p;
	}
}
